from ruamel.yaml.comments import CommentedSeq, CommentedMap
from strictyaml.exceptions import raise_exception
from strictyaml.validators import Validator
from strictyaml.representation import YAML
from strictyaml import constants
from strictyaml import utils
import dateutil.parser
import decimal
import sys
import re


if sys.version_info[0] == 3:
    unicode = str


class Scalar(Validator):
    @property
    def rule_description(self):
        return "a {0}".format(self.__class__.__name__.lower())

    def validate(self, chunk):
        val = chunk.contents

        if type(val) == CommentedSeq or type(val) == CommentedMap:
            raise_exception(
                "when expecting a {0}".format(self.__class__.__name__.lower()),
                "found mapping/sequence",
                chunk,
            )
        else:
            return self.validate_scalar(chunk, value=None)


class Enum(Scalar):
    def __init__(self, restricted_to):
        for element in restricted_to:
            assert type(element) is str
        self._restricted_to = restricted_to

    def validate_scalar(self, chunk, value):
        val = unicode(chunk.contents) if value is None else value
        if val not in self._restricted_to:
            raise_exception(
                "when expecting one of: {0}".format(", ".join(self._restricted_to)),
                "found '{0}'".format(val),
                chunk,
            )
        else:
            return YAML(val, chunk=chunk, validator=self)

    def __repr__(self):
        return u"Enum({0})".format(repr(self._restricted_to))


class CommaSeparated(Scalar):
    def __init__(self, item_validator):
        self._item_validator = item_validator

    def validate_scalar(self, chunk, value):
        val = unicode(chunk.contents) if value is None else value
        return YAML(
            [
                YAML(self._item_validator.validate_scalar(chunk, value=item.lstrip()))
                for item in val.split(",")
            ],
            chunk=chunk,
            validator=self,
        )

    def __repr__(self):
        return "CommaSeparated({0})".format(self._item_validator)


class Regex(Scalar):
    def __init__(self, regular_expression):
        """
        Give regular expression, e.g. u'[0-9]'
        """
        self._regex = regular_expression
        self._matching_message = "when expecting string matching {0}".format(self._regex)

    def validate_scalar(self, chunk, value=None):
        if re.compile(self._regex).match(chunk.contents) is None:
            raise_exception(
                self._matching_message,
                "found non-matching string",
                chunk,
            )
        return YAML(
            unicode(chunk.contents) if value is None else value,
            text=chunk.contents,
            chunk=chunk,
            validator=self,
        )


class Email(Regex):
    def __init__(self):
        self._regex = constants.REGEXES['email']
        self._matching_message = "when expecting an email address"


class Url(Regex):
    def __init__(self):
        self._regex = constants.REGEXES['url']
        self._matching_message = "when expecting a url"


class Str(Scalar):
    def validate_scalar(self, chunk, value=None):
        return YAML(
            unicode(chunk.contents) if value is None else value,
            text=chunk.contents,
            chunk=chunk,
            validator=self,
        )

    def __repr__(self):
        return u"Str()"


class Int(Scalar):
    def validate_scalar(self, chunk, value=None):
        val = unicode(chunk.contents) if value is None else value
        if not utils.is_integer(val):
            raise_exception(
                    "when expecting an integer",
                    "found non-integer",
                    chunk,
                )
        else:
            return YAML(int(val), val, chunk=chunk, validator=self)

    def __repr__(self):
        return u"Int()"


class Bool(Scalar):
    def validate_scalar(self, chunk, value=None):
        val = unicode(chunk.contents) if value is None else value
        if unicode(val).lower() not in constants.BOOL_VALUES:
            raise_exception(
                """when expecting a boolean value (one of "{0}")""".format(
                    '", "'.join(constants.BOOL_VALUES)
                ),
                "found non-boolean",
                chunk,
            )
        else:
            if val.lower() in constants.TRUE_VALUES:
                return YAML(True, val, chunk=chunk, validator=self)
            else:
                return YAML(False, val, chunk=chunk, validator=self)

    def __repr__(self):
        return u"Bool()"


class Float(Scalar):
    def validate_scalar(self, chunk, value=None):
        val = unicode(chunk.contents) if value is None else value
        if not utils.is_decimal(val):
            raise_exception(
                "when expecting a float",
                "found non-float",
                chunk,
            )
        else:
            return YAML(float(val), val, chunk=chunk, validator=self)

    def __repr__(self):
        return u"Float()"


class Decimal(Scalar):
    def validate_scalar(self, chunk, value=None):
        val = unicode(chunk.contents) if value is None else value
        if not utils.is_decimal(val):
            raise_exception(
                "when expecting a decimal",
                "found non-decimal",
                chunk,
            )
        else:
            return YAML(decimal.Decimal(val), val, chunk=chunk, validator=self)

    def __repr__(self):
        return u"Decimal()"


class Datetime(Scalar):
    def validate_scalar(self, chunk, value=None):
        val = unicode(chunk.contents) if value is None else value

        try:
            return YAML(
                dateutil.parser.parse(val),
                val,
                chunk=chunk,
                validator=self
            )
        except ValueError:
            raise_exception(
                "when expecting a datetime",
                "found non-datetime",
                chunk,
            )

    def __repr__(self):
        return u"Datetime()"


class EmptyNone(Scalar):
    def validate_scalar(self, chunk, value):
        val = unicode(chunk.contents) if value is None else value
        if val != "":
            raise_exception(
                "when expecting an empty value",
                "found non-empty value",
                chunk,
            )
        else:
            return self.empty(chunk)

    def empty(self, chunk):
        return YAML(None, '', chunk=chunk, validator=self)

    def __repr__(self):
        return u"EmptyNone()"


class EmptyDict(EmptyNone):
    def empty(self, chunk):
        return YAML({}, '', chunk=chunk, validator=self)

    def __repr__(self):
        return u"EmptyDict()"


class EmptyList(EmptyNone):
    def empty(self, chunk):
        return YAML([], '', chunk=chunk, validator=self)

    def __repr__(self):
        return u"EmptyList()"



from ruamel.yaml.comments import CommentedSeq, CommentedMap
from ruamel.yaml import dump, RoundTripDumper
from copy import deepcopy


class YAMLChunk(object):
    def __init__(self, document, pointer=None):
        self._contents = None
        self._document = document
        self._pointer = pointer if pointer is not None \
            else YAMLPointer()

    def index(self, index):
        return YAMLChunk(self._document, self._pointer.index(index))

    def val(self, index):
        return YAMLChunk(self._document, self._pointer.val(index))

    def key(self, name):
        return YAMLChunk(self._document, self._pointer.key(name))

    def start_line(self):
        return self._pointer.start_line(self._document)

    def end_line(self):
        return self._pointer.end_line(self._document)

    def lines(self):
        return self._pointer.lines(self._document)

    def lines_before(self, how_many):
        return self._pointer.lines_before(self._document, how_many)

    def lines_after(self, how_many):
        return self._pointer.lines_after(self._document, how_many)

    @property
    def document(self):
        return self._document

    @property
    def contents(self):
        return self._pointer.get(self._document)


class YAMLPointer(object):
    def __init__(self):
        self._indices = []

    def val(self, index):
        new_location = deepcopy(self)
        new_location._indices.append(('val', index))
        return new_location

    def key(self, name):
        new_location = deepcopy(self)
        new_location._indices.append(('key', name))
        return new_location

    def index(self, index):
        new_location = deepcopy(self)
        new_location._indices.append(('index', index))
        return new_location

    def _slice_segment(self, indices, segment, include_selected):
        slicedpart = deepcopy(segment)

        if len(indices) == 0 and not include_selected:
            slicedpart = None
        else:
            if len(indices) > 0:
                index = indices[0][1]
                start_popping = False

                if type(segment) == CommentedMap:
                    for key in segment.keys():
                        if start_popping:
                            slicedpart.pop(key)

                        if index == key:
                            start_popping = True

                            if type(segment[index]) in (CommentedSeq, CommentedMap, ):
                                slicedpart[index] = self._slice_segment(
                                    indices[1:],
                                    segment[index],
                                    include_selected=include_selected
                                )

                            if not include_selected and len(indices) == 1:
                                slicedpart.pop(key)

                if type(segment) == CommentedSeq:
                    for i, value in enumerate(segment):
                        if start_popping:
                            del slicedpart[-1]

                        if i == index:
                            start_popping = True

                            if type(segment[index]) in (CommentedSeq, CommentedMap, ):
                                slicedpart[index] = self._slice_segment(
                                    indices[1:],
                                    segment[index],
                                    include_selected=include_selected
                                )

                            if not include_selected and len(indices) == 1:
                                slicedpart.pop(index)

        return slicedpart

    def start_line(self, document):
        slicedpart = self._slice_segment(self._indices, document, include_selected=False)

        if slicedpart is None or slicedpart == {} or slicedpart == []:
            return 1
        else:
            return len(dump(slicedpart, Dumper=RoundTripDumper).rstrip().split('\n')) + 1

    def end_line(self, document):
        slicedpart = self._slice_segment(self._indices, document, include_selected=True)
        return len(dump(slicedpart, Dumper=RoundTripDumper).rstrip().split('\n'))

    def lines(self, document):
        return "\n".join(dump(document, Dumper=RoundTripDumper).split('\n')[
            self.start_line(document) - 1:self.end_line(document)
        ])

    def lines_before(self, document, how_many):
        return "\n".join(dump(document, Dumper=RoundTripDumper).split('\n')[
            self.start_line(document) - 1 - how_many:self.start_line(document) - 1
        ])

    def lines_after(self, document, how_many):
        return "\n".join(dump(document, Dumper=RoundTripDumper).split('\n')[
            self.end_line(document):self.end_line(document) + how_many
        ])

    def get(self, document):
        segment = document
        for index_type, index in self._indices:
            if index_type == "val":
                segment = segment[index]
            elif index_type == "index":
                segment = segment[index]
            else:
                segment = index
        return deepcopy(segment)

    def __repr__(self):
        return "<YAMLPointer: {0}>".format(self._indices)



from re import compile


def is_integer(value):
    return compile("^[-+]?\d+$").match(value) is not None


def is_decimal(value):
    return compile(r"^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$").match(value) is not None



from ruamel.yaml.comments import CommentedSeq, CommentedMap
from strictyaml.exceptions import YAMLValidationError
from strictyaml.exceptions import raise_exception
from strictyaml.representation import YAML
import sys

if sys.version_info[0] == 3:
    unicode = str


class Optional(object):
    def __init__(self, key):
        self.key = key


class Validator(object):
    def __or__(self, other):
        return OrValidator(self, other)

    def __call__(self, chunk):
        return self.validate(chunk)


class OrValidator(Validator):
    def __init__(self, validator_a, validator_b):
        self._validator_a = validator_a
        self._validator_b = validator_b

    def validate(self, chunk):
        try:
            return self._validator_a(chunk)
        except YAMLValidationError:
            return self._validator_b(chunk)

    def __repr__(self):
        return u"{0} | {1}".format(repr(self._validator_a), repr(self._validator_b))


class MapPattern(Validator):
    def __init__(self, key_validator, value_validator):
        self._key_validator = key_validator
        self._value_validator = value_validator

    def validate(self, chunk):
        return_snippet = chunk.contents

        if not isinstance(return_snippet, CommentedMap):
            raise_exception(
                "when expecting a mapping",
                "found non-mapping",
                chunk,
            )
        else:
            for key, value in chunk.contents.items():
                valid_key = self._key_validator(chunk.key(key))
                valid_val = self._value_validator(chunk.val(key))
                return_snippet[valid_key] = valid_val

                del return_snippet[valid_key]
                return_snippet[valid_key] = self._value_validator(
                    chunk.val(key)
                )

        return YAML(return_snippet, chunk=chunk, validator=self)

    def __repr__(self):
        return u"MapPattern({0}, {1})".format(
            repr(self._key_validator), repr(self._value_validator)
        )


class Map(Validator):
    def __init__(self, validator):
        self._validator = validator

        self._validator_dict = {
            key.key if type(key) == Optional else key: value for key, value in validator.items()
        }

    def __repr__(self):
        return u"Map({{{0}}})".format(', '.join([
            '{0}: {1}'.format(
                'Optional("{0}")'.format(key.key) if type(key) is Optional else '"{0}"'.format(key),
                repr(value)
            ) for key, value in self._validator.items()
        ]))

    def validate(self, chunk):
        return_snippet = chunk.contents

        if type(chunk.contents) != CommentedMap:
            raise_exception(
                "when expecting a mapping",
                "found non-mapping",
                chunk,
            )
        else:
            for key, value in chunk.contents.items():
                if key not in self._validator_dict.keys():
                    raise_exception(
                        u"while parsing a mapping",
                        u"unexpected key not in schema '{0}'".format(unicode(key)),
                        chunk.key(key)
                    )

                validator = self._validator_dict[key]
                del return_snippet[key]
                return_snippet[
                    YAML(key, chunk=chunk.key(key), validator=validator)
                ] = validator(chunk.val(key))

        return YAML(return_snippet, chunk=chunk, validator=self)


class Seq(Validator):
    def __init__(self, validator):
        self._validator = validator

    def __repr__(self):
        return "Seq({0})".format(repr(self._validator))

    def validate(self, chunk):
        return_snippet = chunk.contents

        if not isinstance(return_snippet, CommentedSeq):
            raise_exception(
                "when expecting a sequence",
                "found non-sequence",
                chunk,
            )
        else:
            for i, item in enumerate(chunk.contents):
                return_snippet[i] = self._validator(chunk.index(i))

        return YAML(return_snippet, chunk=chunk, validator=self)


class FixedSeq(Validator):
    def __init__(self, validators):
        self._validators = validators

    def __repr__(self):
        return "FixedSeq({0})".format(repr(self._validators))

    def validate(self, chunk):
        return_snippet = chunk.contents

        if not isinstance(return_snippet, CommentedSeq):
            raise_exception(
                "when expecting a sequence of {0} elements".format(len(self._validators)),
                "found non-sequence",
                chunk,
            )
        else:
            if len(self._validators) != len(chunk.contents):
                raise_exception(
                    "when expecting a sequence of {0} elements".format(len(self._validators)),
                    "found a sequence of {0} elements".format(len(chunk.contents)),
                    chunk,
                )
            for i, item_and_val in enumerate(zip(chunk.contents, self._validators)):
                item, validator = item_and_val
                return_snippet[i] = validator(chunk.index(i))

        return YAML(return_snippet, chunk=chunk, validator=self)


class UniqueSeq(Validator):
    def __init__(self, validator):
        self._validator = validator

    def __repr__(self):
        return "UniqueSeq({0})".format(repr(self._validator))

    def validate(self, chunk):
        return_snippet = chunk.contents

        if type(chunk.contents) != CommentedSeq:
            raise_exception(
                "when expecting a unique sequence",
                "found non-sequence",
                chunk,
            )
        else:
            existing_items = set()

            for i, item in enumerate(chunk.contents):
                if item in existing_items:
                    raise_exception(
                        "while parsing a sequence",
                        "duplicate found",
                        chunk
                    )
                else:
                    existing_items.add(item)
                    return_snippet[i] = self._validator(chunk.index(i))

        return YAML(return_snippet, chunk=chunk, validator=self)



# The all important loader
from strictyaml.parser import load

# YAML object
from strictyaml.representation import YAML

# Validators
from strictyaml.validators import Optional
from strictyaml.validators import Validator
from strictyaml.validators import OrValidator
from strictyaml.any_validator import Any
from strictyaml.scalar import Scalar
from strictyaml.scalar import Enum
from strictyaml.scalar import Regex
from strictyaml.scalar import Email
from strictyaml.scalar import Url
from strictyaml.scalar import Str
from strictyaml.scalar import Int
from strictyaml.scalar import Bool
from strictyaml.scalar import Float
from strictyaml.scalar import Decimal
from strictyaml.scalar import Datetime
from strictyaml.scalar import CommaSeparated
from strictyaml.scalar import EmptyNone
from strictyaml.scalar import EmptyDict
from strictyaml.scalar import EmptyList
from strictyaml.validators import Map
from strictyaml.validators import MapPattern
from strictyaml.validators import Seq
from strictyaml.validators import UniqueSeq
from strictyaml.validators import FixedSeq

# Base exception from ruamel.yaml (all exceptions inherit from this)
from ruamel.yaml import YAMLError

# Exceptions
from strictyaml.exceptions import StrictYAMLError
from strictyaml.exceptions import YAMLValidationError

# Disallowed token exceptions
from strictyaml.exceptions import DisallowedToken

from strictyaml.exceptions import TagTokenDisallowed
from strictyaml.exceptions import FlowMappingDisallowed
from strictyaml.exceptions import AnchorTokenDisallowed
from strictyaml.exceptions import DuplicateKeysDisallowed
from strictyaml import exceptions


__version__ = "0.7.0"
